@@ -30,6 +30,7 @@
 #include "knotification.h"
 #include "debug_p.h"
 
+#include <QApplication>
 #include <QBuffer>
 #include <QImage>
 #include <QLabel>
@@ -52,6 +53,7 @@
 #include <QScreen>
 #include <QFontMetrics>
 #include <QIcon>
+#include <QUrl>
 
 #include <kconfiggroup.h>
 #include <KCodecs/KCharsets>
@@ -107,6 +109,9 @@ public:
      * Specifies if DBus Notifications interface exists on session bus
      */
     bool dbusServiceExists;
+
+    bool dbusServiceActivatable;
+
     /**
      * DBus notification daemon capabilities cache.
      * Do not use this variable. Use #popupServerCapabilities() instead.
@@ -159,53 +164,40 @@ NotifyByPopup::NotifyByPopup(QObject *parent)
 {
     d->animationTimer = 0;
     d->dbusServiceExists = false;
+    d->dbusServiceActivatable = false;
     d->dbusServiceCapCacheDirty = true;
     d->nextPosition = -1;
 
     // check if service already exists on plugin instantiation
     QDBusConnectionInterface *interface = QDBusConnection::sessionBus().interface();
-    d->dbusServiceExists = interface && interface->isServiceRegistered(dbusServiceName);
+    d->dbusServiceExists = interface && interface->isServiceRegistered(QString::fromLatin1(dbusServiceName));
 
     if (d->dbusServiceExists) {
-        onServiceOwnerChanged(dbusServiceName, QString(), QStringLiteral("_")); //connect signals
+        onServiceOwnerChanged(QString::fromLatin1(dbusServiceName), QString(), QStringLiteral("_")); //connect signals
     }
 
     // to catch register/unregister events from service in runtime
     QDBusServiceWatcher *watcher = new QDBusServiceWatcher(this);
     watcher->setConnection(QDBusConnection::sessionBus());
     watcher->setWatchMode(QDBusServiceWatcher::WatchForOwnerChange);
-    watcher->addWatchedService(dbusServiceName);
+    watcher->addWatchedService(QString::fromLatin1(dbusServiceName));
     connect(watcher, SIGNAL(serviceOwnerChanged(QString,QString,QString)),
             SLOT(onServiceOwnerChanged(QString,QString,QString)));
 
+#ifndef Q_WS_WIN
     if (!d->dbusServiceExists) {
-        bool startfdo = false;
-#ifdef Q_WS_WIN
-        startfdo = true;
-#else
-        if (qEnvironmentVariableIsEmpty("KDE_FULL_SESSION")) {
-            QDBusMessage message = QDBusMessage::createMethodCall(QStringLiteral("org.freedesktop.DBus"),
-                                                                  QStringLiteral("/org/freedesktop/DBus"),
-                                                                  QStringLiteral("org.freedesktop.DBus"),
-                                                                  QStringLiteral("ListActivatableNames"));
-
-            // FIXME - this should be async
-            QDBusReply<QStringList> reply = QDBusConnection::sessionBus().call(message);
-            if (reply.isValid() && reply.value().contains(dbusServiceName)) {
-                startfdo = true;
-                // We need to set d->dbusServiceExists to true because dbus might be too slow
-                // starting the service and the first call to NotifyByPopup::notify
-                // might not have had the service up, by setting this to true we
-                // guarantee it will still go through dbus and dbus will do the correct
-                // thing and wait for the service to go up
-                d->dbusServiceExists = true;
-            }
-        }
-#endif
-        if (startfdo) {
-            QDBusConnection::sessionBus().interface()->startService(dbusServiceName);
+        QDBusMessage message = QDBusMessage::createMethodCall(QStringLiteral("org.freedesktop.DBus"),
+                                                                QStringLiteral("/org/freedesktop/DBus"),
+                                                                QStringLiteral("org.freedesktop.DBus"),
+                                                                QStringLiteral("ListActivatableNames"));
+        QDBusReply<QStringList> reply = QDBusConnection::sessionBus().call(message);
+        if (reply.isValid() && reply.value().contains(QString::fromLatin1(dbusServiceName))) {
+            d->dbusServiceActivatable = true;
+            //if the service is activatable, we can assume it exists even if it is not currently running
+            d->dbusServiceExists = true;
         }
     }
+#endif
 }
 
 
@@ -249,7 +241,7 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
 
     // Persistent     => 0  == infinite timeout
     // CloseOnTimeout => -1 == let the server decide
-    int timeout = notification->flags() & KNotification::Persistent ? 0 : -1;
+    int timeout = (notification->flags() & KNotification::Persistent) ? 0 : -1;
 
     // if Growl can display our popups, use that instead
     if (NotifyByPopupGrowl::canPopup()) {
@@ -270,7 +262,7 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
 
         // The first arg is QPixmap*, however that pixmap is not used
         // at all (it has Q_UNUSED) so just set it to 0
-        NotifyByPopupGrowl::popup(Q_NULLPTR, timeout, appCaption, text);
+        NotifyByPopupGrowl::popup(nullptr, timeout, appCaption, text);
 
         // Finish immediately, because current NotifyByPopupGrowl can't callback
         finish(notification);
@@ -284,6 +276,11 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
         return;
     }
 
+    if (!qobject_cast<QApplication *>(QCoreApplication::instance())) {
+        qCWarning(LOG_KNOTIFICATIONS) << "KNotification did not detect any running org.freedesktop.Notifications server and fallback notifications cannot be used without a QApplication!";
+        return;
+    }
+
     // last fallback - display the popup using KPassivePopup
     KPassivePopup *pop = new KPassivePopup(notification->widget());
     d->passivePopups.insert(notification, pop);
@@ -310,8 +307,7 @@ void NotifyByPopup::onPassivePopupDestroyed()
         return;
     }
 
-    QMap<KNotification*, KPassivePopup*>::iterator it;
-    for (it = d->passivePopups.begin(); it != d->passivePopups.end(); ++it) {
+    for (QMap<KNotification*, KPassivePopup*>::iterator it = d->passivePopups.begin(); it != d->passivePopups.end(); ++it) {
         QObject *popup = it.value();
         if (popup && popup == destroyedPopup) {
             finish(it.key());
@@ -357,8 +353,8 @@ void NotifyByPopup::timerEvent(QTimerEvent *event)
 
 void NotifyByPopup::onPassivePopupLinkClicked(const QString &link)
 {
-    unsigned int id = link.section('/' , 0 , 0).toUInt();
-    unsigned int action = link.section('/' , 1 , 1).toUInt();
+    unsigned int id = link.section(QLatin1Char('/') , 0 , 0).toUInt();
+    unsigned int action = link.section(QLatin1Char('/') , 1 , 1).toUInt();
 
     if (id == 0 || action == 0) {
         return;
@@ -431,17 +427,19 @@ void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QStr
 
     d->dbusServiceCapCacheDirty = true;
     d->popupServerCapabilities.clear();
-    d->notificationQueue.clear();
 
     if (newOwner.isEmpty()) {
-        d->dbusServiceExists = false;
+        d->notificationQueue.clear();
+        if (!d->dbusServiceActivatable) {
+            d->dbusServiceExists = false;
+        }
     } else if (oldOwner.isEmpty()) {
         d->dbusServiceExists = true;
 
         // connect to action invocation signals
         bool connected = QDBusConnection::sessionBus().connect(QString(), // from any service
-                                                               dbusPath,
-                                                               dbusInterfaceName,
+                                                               QString::fromLatin1(dbusPath),
+                                                               QString::fromLatin1(dbusInterfaceName),
                                                                QStringLiteral("ActionInvoked"),
                                                                this,
                                                                SLOT(onGalagoNotificationActionInvoked(uint,QString)));
@@ -450,8 +448,8 @@ void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QStr
         }
 
         connected = QDBusConnection::sessionBus().connect(QString(), // from any service
-                                                          dbusPath,
-                                                          dbusInterfaceName,
+                                                          QString::fromLatin1(dbusPath),
+                                                          QString::fromLatin1(dbusInterfaceName),
                                                           QStringLiteral("NotificationClosed"),
                                                           this,
                                                           SLOT(onGalagoNotificationClosed(uint,uint)));
@@ -470,7 +468,11 @@ void NotifyByPopup::onGalagoNotificationActionInvoked(uint notificationId, const
 
     KNotification *n = *iter;
     if (n) {
-        emit actionInvoked(n->id(), actionKey.toUInt());
+        if (actionKey == QStringLiteral("default")) {
+            emit actionInvoked(n->id(), 0);
+        } else {
+            emit actionInvoked(n->id(), actionKey.toUInt());
+        }
     } else {
         d->galagoNotifications.erase(iter);
     }
@@ -519,7 +521,7 @@ void NotifyByPopup::onGalagoServerCapabilitiesReceived(const QStringList &capabi
     d->dbusServiceCapCacheDirty = false;
 
     // re-run notify() on all enqueued notifications
-    for (int i = 0; i < d->notificationQueue.size(); i++) {
+    for (int i = 0, total = d->notificationQueue.size(); i < total; ++i) {
         notify(d->notificationQueue.at(i).first, d->notificationQueue.at(i).second);
     }
 
@@ -621,7 +623,7 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         }
     }
 
-    QDBusMessage dbusNotificationMessage = QDBusMessage::createMethodCall(dbusServiceName, dbusPath, dbusInterfaceName, QStringLiteral("Notify"));
+    QDBusMessage dbusNotificationMessage = QDBusMessage::createMethodCall(QString::fromLatin1(dbusServiceName), QString::fromLatin1(dbusPath), QString::fromLatin1(dbusInterfaceName), QStringLiteral("Notify"));
 
     QList<QVariant> args;
 
@@ -660,6 +662,11 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
     // (i.e. starting from 1)
     QStringList actionList;
     if (popupServerCapabilities.contains(QStringLiteral("actions"))) {
+        QString defaultAction = notification->defaultAction();
+        if (!defaultAction.isEmpty()) {
+            actionList.append(QStringLiteral("default"));
+            actionList.append(defaultAction);
+        }
         int actId = 0;
         Q_FOREACH (const QString &actionName, notification->actions()) {
             actId++;
@@ -686,6 +693,18 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         hintsMap[QStringLiteral("x-kde-skipGrouping")] = 1;
     }
 
+    if (!notification->urls().isEmpty()) {
+        hintsMap[QStringLiteral("x-kde-urls")] = QUrl::toStringList(notification->urls());
+    }
+
+    if (!(notification->flags() & KNotification::Persistent)) {
+        hintsMap[QStringLiteral("transient")] = true;
+    }
+
+    if (!qApp->desktopFileName().isEmpty()) {
+        hintsMap[QStringLiteral("desktop-entry")] = qApp->desktopFileName();
+    }
+
     //FIXME - reenable/fix
     // let's see if we've got an image, and store the image in the hints map
     if (!notification->pixmap().isNull()) {
@@ -701,7 +720,7 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
 
     // Persistent     => 0  == infinite timeout
     // CloseOnTimeout => -1 == let the server decide
-    int timeout = notification->flags() & KNotification::Persistent ? 0 : -1;
+    int timeout = (notification->flags() & KNotification::Persistent) ? 0 : -1;
 
     args.append(timeout); // expire timout
 
@@ -728,8 +747,8 @@ void NotifyByPopupPrivate::closeGalagoNotification(KNotification *notification)
         return;
     }
 
-    QDBusMessage m = QDBusMessage::createMethodCall(dbusServiceName, dbusPath,
-                                                    dbusInterfaceName, QStringLiteral("CloseNotification"));
+    QDBusMessage m = QDBusMessage::createMethodCall(QString::fromLatin1(dbusServiceName), QString::fromLatin1(dbusPath),
+                                                    QString::fromLatin1(dbusInterfaceName), QStringLiteral("CloseNotification"));
     QList<QVariant> args;
     args.append(galagoId);
     m.setArguments(args);
@@ -755,22 +774,22 @@ void NotifyByPopupPrivate::queryPopupServerCapabilities()
     }
 
     if (dbusServiceCapCacheDirty) {
-        QDBusMessage m = QDBusMessage::createMethodCall(dbusServiceName,
-                                                        dbusPath,
-                                                        dbusInterfaceName,
+        QDBusMessage m = QDBusMessage::createMethodCall(QString::fromLatin1(dbusServiceName),
+                                                        QString::fromLatin1(dbusPath),
+                                                        QString::fromLatin1(dbusInterfaceName),
                                                         QStringLiteral("GetCapabilities"));
 
         QDBusConnection::sessionBus().callWithCallback(m,
                                                        q,
                                                        SLOT(onGalagoServerCapabilitiesReceived(QStringList)),
-                                                       0,
+                                                       nullptr,
                                                        -1);
     }
 }
 
 QString NotifyByPopupPrivate::stripHtml(const QString &text)
 {
-    QXmlStreamReader r("<elem>" + text + "</elem>");
+    QXmlStreamReader r(QStringLiteral("<elem>") + text + QStringLiteral("</elem>"));
     HtmlEntityResolver resolver;
     r.setEntityResolver(&resolver);
     QString result;
@@ -778,8 +797,8 @@ QString NotifyByPopupPrivate::stripHtml(const QString &text)
         r.readNext();
         if (r.tokenType() == QXmlStreamReader::Characters) {
             result.append(r.text());
-        } else if (r.tokenType() == QXmlStreamReader::StartElement && r.name() == "br") {
-            result.append("\n");
+        } else if (r.tokenType() == QXmlStreamReader::StartElement && r.name() == QLatin1String("br")) {
+            result.append(QLatin1Char('\n'));
         }
     }
 
@@ -803,13 +822,13 @@ QString NotifyByPopupPrivate::HtmlEntityResolver::resolveUndeclaredEntity(const
         return result;
     }
 
-    QChar ent = KCharsets::fromEntity('&' + name);
+    QChar ent = KCharsets::fromEntity(QLatin1Char('&') + name);
 
     if (ent.isNull()) {
         qCWarning(LOG_KNOTIFICATIONS) << "Notification to send to backend which does "
                       "not support HTML, contains invalid entity: "
                    << name;
-        ent = ' ';
+        ent = QLatin1Char(' ');
     }
 
     return QString(ent);
